/**
* @file rngTest.c
* @brief FIPS 140-2, section 4.9.1 test implementation
* @author Kosenkov
* @date 23.05.2018
*/

#include "rngTest.h"
#include "memory.h"

/* FIPS 140-2, section 4.9.1 */
#define MONOBIT_MIN     ((uint16_t)9725)
#define MONOBIT_MAX     ((uint16_t)10275)

#define POKER_MIN       ((uint32_t)1563175) /*< formula: 2.16 < 16 / 50 * SUM - 5000 < 46.17 */
#define POKER_MAX       ((uint32_t)1579929)

#define RUNS1_MIN       ((uint16_t)2343)
#define RUNS1_MAX       ((uint16_t)2657)
#define RUNS2_MIN       ((uint16_t)1135)
#define RUNS2_MAX       ((uint16_t)1365)
#define RUNS3_MIN       ((uint16_t)542)
#define RUNS3_MAX       ((uint16_t)708)
#define RUNS4_MIN       ((uint16_t)251)
#define RUNS4_MAX       ((uint16_t)373)
#define RUNS5_MIN       ((uint16_t)111)
#define RUNS5_MAX       ((uint16_t)201)
#define RUNS6_MIN       ((uint16_t)111)
#define RUNS6_MAX       ((uint16_t)201)

#define LONG_RUNS_MAX   ((uint8_t)26)

uint16_t gBytesCnt __attribute__((section(".bss")));

uint16_t gMonobitOnes __attribute__((section(".bss")));

uint16_t gPokerTetradesCnt[16] __attribute__((section(".bss")));

const uint16_t gRunsDesiredMin[7] = {0, RUNS1_MIN, RUNS2_MIN, RUNS3_MIN, RUNS4_MIN, RUNS5_MIN, RUNS6_MIN};
const uint16_t gRunsDesiredMax[7] = {1, RUNS1_MAX, RUNS2_MAX, RUNS3_MAX, RUNS4_MAX, RUNS5_MAX, RUNS6_MAX};
uint16_t gRunsCnts[2][7] __attribute__((section(".bss")));
uint8_t  gRunsPrevBit __attribute__((section(".bss")));
uint16_t gRunsPrevCnt __attribute__((section(".bss")));

uint8_t  gLongRunsPrevBit __attribute__((section(".bss")));
uint16_t gLongRunsCnt __attribute__((section(".bss")));

/**
* Monobit testing step
* 
* @param byte Byte for testing
* @author Kosenkov
* @date 05.06.2018
*/
void monobitTestStep(uint8_t byte)
{
    uint8_t i;
    for (i = 0; i < 8; ++i)
    {
        gMonobitOnes += ((byte >> i) & 1);
    }
}

/**
* Gets result after monobit testing
* 
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 05.06.2018
*/
CM0_STATUS monobitTestResult()
{
    uint16_t res = gMonobitOnes;
    gMonobitOnes = 0;
    return ((MONOBIT_MIN < res && MONOBIT_MAX > res) ? CM0_OK : CM0_FAIL);
}

/**
* Poker testing step (Chi2)
* 
* @param byte Byte for testing
* @author Kosenkov
* @date 05.06.2018
*/
void pokerTestStep(uint8_t byte)
{
    ++gPokerTetradesCnt[byte & 0x0F];
    ++gPokerTetradesCnt[byte >> 4];
}

/**
* Gets result after poker testing
* 
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 05.06.2018
*/
CM0_STATUS pokerTestResult()
{
    uint8_t i;
    uint32_t sum = 0;

    for (i = 0; i < 16; ++i)
    {
        sum += ((uint32_t)gPokerTetradesCnt[i] * (uint32_t)gPokerTetradesCnt[i]);
    }
    mldrMemSet(gPokerTetradesCnt, 0, sizeof(gPokerTetradesCnt));

    return ((POKER_MIN < sum && POKER_MAX > sum) ? CM0_OK : CM0_FAIL);
}

/**
* Runs testing step
* 
* @param byte Byte for testing
* @author Kosenkov
* @date 05.06.2018
*/
void runsTestStep(uint8_t byte)
{
    uint8_t i;
    uint8_t curBit;

    for (i = 0; i < 8; ++i)
    {
        curBit = (byte >> i) & 1;
        if (gRunsPrevBit == curBit)
        {
            ++gRunsPrevCnt;
        }
        else
        {
            gRunsPrevBit = curBit;
            ++gRunsCnts[gRunsPrevBit][gRunsPrevCnt > 5 ? 6 : gRunsPrevCnt];
            gRunsPrevCnt = 1;
        }
    }
}

/**
* Gets result after runs testing
* 
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 05.06.2018
*/
CM0_STATUS runsTestResult()
{
    uint8_t i;
    CM0_STATUS res = CM0_OK;

    for (i = 1; i < 7; ++i)
    {
        if (gRunsCnts[0][i] < gRunsDesiredMin[i] || gRunsCnts[0][i] > gRunsDesiredMax[i])
        {
            res = CM0_FAIL;
            break;
        }
        if (gRunsCnts[1][i] < gRunsDesiredMin[i] || gRunsCnts[1][i] > gRunsDesiredMax[i])
        {
            res = CM0_FAIL;
            break;
        }
    }
    mldrMemSet(gRunsCnts, 0, sizeof(gRunsCnts));
    gRunsPrevBit = 0xFF;
    gRunsPrevCnt = 0;

    return res;
}

/**
* Long runs testing step
* 
* @param byte Byte for testing
* @author Kosenkov
* @date 05.06.2018
*/
void longRunsTestStep(uint8_t byte)
{
    uint8_t i;
    uint8_t curBit;

    if (gLongRunsCnt >= LONG_RUNS_MAX)
    {
        return;
    }

    for (i = 0; i < 8; ++i)
    {
        curBit = (byte >> i) & 1;
        if (gLongRunsPrevBit == curBit)
        {
            ++gLongRunsCnt;
        }
        else
        {
            gLongRunsPrevBit = curBit;
            if (gLongRunsCnt >= LONG_RUNS_MAX)
            {
                return;
            }
            else
            {
                gLongRunsCnt = 1;
            }
        }
    }
}

/**
* Gets result after long runs testing
* 
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 05.06.2018
*/
CM0_STATUS longRunsTestResult()
{
    uint16_t cnt = gLongRunsCnt;
    gLongRunsCnt = 0;
    gLongRunsPrevBit = 0xFF;
    return ((cnt < LONG_RUNS_MAX) ? CM0_OK : CM0_FAIL);
}

/**
* Initialize read-write global variables inside the code, which work
* after scramber initialization
*
* @author Kosenkov
* @date 25.07.2018
*/
void rngTestInit()
{
    gBytesCnt = 0;
    gMonobitOnes = 0;
    gRunsPrevBit = 0xFF;
    gRunsPrevCnt = 0;
    gLongRunsPrevBit = 0xFF;
    gLongRunsCnt = 0;
    mldrMemSet(gPokerTetradesCnt, 0, sizeof(gPokerTetradesCnt));
    mldrMemSet(gRunsCnts, 0, sizeof(gRunsCnts));
}

/**
* The function randomTest performs testing step of FIPS 140-2, section 4.9.1 of 
* byte
* 
* @param byte Random byte that should be tested
* @author Kosenkov
* @date 05.06.2018
*/
void rngTestByte(uint8_t byte)
{
    ++gBytesCnt;
    if (STAT_CNT < gBytesCnt)
    {
        --gBytesCnt;
        return;
    }
    
    monobitTestStep(byte);
    pokerTestStep(byte);
    runsTestStep(byte);
    longRunsTestStep(byte);
}

/**
* Check results of all tests
* 
* @retval CM0_OK if success
* @retval CM0_RNG_TEST_ERR otherwise
* @author Kosenkov
* @date 05.06.2018
*/
CM0_STATUS rngTestResult()
{
    CM0_STATUS monobitRes;
    CM0_STATUS pokerRes;
    CM0_STATUS runsRes;
    CM0_STATUS longRunsRes;
    
    if (STAT_CNT != gBytesCnt)
    {
        return CM0_RNG_TEST_ERR;
    }
    
    gBytesCnt = 0;
    
    monobitRes = monobitTestResult();
    pokerRes = pokerTestResult();
    runsRes = runsTestResult();
    longRunsRes = longRunsTestResult();
    
    if (CM0_CHK_ERR(monobitRes) ||
        CM0_CHK_ERR(pokerRes)   ||
        CM0_CHK_ERR(runsRes)    ||
        CM0_CHK_ERR(longRunsRes))
    {
        return CM0_RNG_TEST_ERR;
    }
    else
    {
        return CM0_OK;
    }
}
