#include "cm0.h"
#include "alu.h"

const Reg128Num gMulReg =    112; /*< Address of mul register as operand */
const Reg128Num gMulRegRes =   0; /*< Address of mul register as result  */
const Reg128Num gZeroReg =   127; /*< Address of zero register           */
const uint8_t   gStackSz =    16; /*< Size of register stack             */

/**
* Data for locating command in ALU bank
*/
typedef struct
{
    uint8_t  Bank;
    uint32_t Beg;
    uint32_t Cnt;
} CmdPtr;

/**
* Command names
*/
typedef enum
{
    ALU_Zero = 0,
    ALU_Push,
    ALU_Pop,
    ALU_Mov,
    ALU_CMov,
    ALU_Add,
    ALU_Sub,
    ALU_Shl,
    ALU_CShl,
    ALU_Shr,
    ALU_CShr,
    ALU_Mul,
    CMD_MAX_I
} Command;

const uint32_t gMaxRegSz = 32; /*< Max size of complex data in register units */
CmdPtr gCmds[CMD_MAX_I];       /*< Array of commands locating data */

////////////////////////////////////////////////////////////////////////////////
// GENERATE COMMANDS
////////////////////////////////////////////////////////////////////////////////
uint32_t gCmd[1024]; /*< Complex command buffer */

/**
* Adds new command to ALU bank
*
* @param name Of command
* @param cmd Array of native commands
* @author Kosenkov
* @date 18.05.2018
*/
void addCmdToAlu(Command name, uint32_t* cmd)
{
    writeBy32(ALU_MEM + gCmds[name].Beg, cmd, gCmds[name].Cnt);
}

/**
* Generation of complex command
* Zeroe registers command
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdZero()
{
    uint8_t  i;

    for (i = 0; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | (gZeroReg << 8) | MOV_CMD;
    }
    
    addCmdToAlu(ALU_Zero, gCmd);
}

/**
* Generation of complex command
* Move register to stack
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdPush()
{
    uint8_t  i;

    // initial bases [][Res][B][A] where Res -> Stack; A -> B; B == Res
    for (i = 0; i < gStackSz; ++i)
    {
        gCmd[i] = (i << 24) | (i << 16) | (i << 8) | PUSH_CMD;
    }
    
    addCmdToAlu(ALU_Push, gCmd);
}

/**
* Generation of complex command
* Move data from stack to register
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdPop()
{
    uint8_t  i;

    // initial bases [][Res][][] where Res <- Stack
    for (i = 0; i < gStackSz; ++i)
    {
        gCmd[i] = ((gStackSz - 1 - i) << 24) | POP_CMD;
    }
    
    addCmdToAlu(ALU_Pop, gCmd);
}

/**
* Generation of complex command
* Move command
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdMov()
{
    uint8_t  i;

    for (i = 0; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | (i << 8) | MOV_CMD;
    }
    
    addCmdToAlu(ALU_Mov, gCmd);
}

/**
* Generation of complex command
* Move with carry
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdCMov()
{
    uint8_t  i;
    
    for (i = 0; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | (i << 8) | C_MOV_CMD;
    }
    
    addCmdToAlu(ALU_CMov, gCmd);
}

/**
* Generation of complex command
* Addition
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdAdd()
{
    uint8_t  i;

    gCmd[0] = ADD_CMD;
    for (i = 1; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | (i << 16) | (i << 8) | C_ADD_CMD;
    }
    
    addCmdToAlu(ALU_Add, gCmd);
}

/**
* Generation of complex command
* Subtraction
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdSub()
{
    uint8_t  i;

    gCmd[0] = 0x00800000 | O_ADD_CMD;
    for (i = 1; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | ((0x80 | i) << 16) | (i << 8) | C_ADD_CMD;
    }
    
    addCmdToAlu(ALU_Sub, gCmd);
}

/**
* Generation of complex command
* Left shift
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdShl()
{
    uint8_t  i;

    gCmd[0] = SHL_CMD;
    for (i = 1; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | (i << 8) | S_SHL_CMD;
    }
    
    addCmdToAlu(ALU_Shl, gCmd);
}

/**
* Generation of complex command
* Left shift with carry
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdCShl()
{
    uint8_t  i;

    gCmd[0] = C_SHL_CMD;
    for (i = 1; i < gMaxRegSz; ++i)
    {
        gCmd[i] = (i << 24) | (i << 8) | S_SHL_CMD;
    }
    
    addCmdToAlu(ALU_CShl, gCmd);
}

/**
* Generation of complex command
* Right shift
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdShr()
{
    uint8_t  i;

    gCmd[0] = ((gMaxRegSz - 1) << 24) | ((gMaxRegSz - 1) << 8) | SHR_CMD;
    for (i = 1; i < gMaxRegSz; ++i)
    {
        gCmd[i] = ((gMaxRegSz - 1 - i) << 24) | ((gMaxRegSz - 1 - i) << 8) | S_SHR_CMD;
    }
    
    addCmdToAlu(ALU_Shr, gCmd);
}

/**
* Generation of complex command
* Right shift with carry
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdCShr()
{
    uint8_t  i;

    gCmd[0] = ((gMaxRegSz - 1) << 24) | ((gMaxRegSz - 1) << 8) | C_SHR_CMD;
    for (i = 1; i < gMaxRegSz; ++i)
    {
        gCmd[i] = ((gMaxRegSz - 1 - i) << 24) | ((gMaxRegSz - 1 - i) << 8) | S_SHR_CMD;
    }
    
    addCmdToAlu(ALU_CShr, gCmd);
}

/**
* Generation of complex command
* Multiplication
*
* @author Kosenkov
* @date 18.05.2018
*/
void genCmdMul()
{
    uint8_t  i;
    
    // initial BASE_REG = 0x[T][C][B][A] with current offsets
    for (i = 0; i < gMaxRegSz; ++i)
    {
        gCmd[12 * i] = 0x04010000 | B_SHFT_CMD; // load [0][B][A]
        gCmd[12 * i + 1] = (gMulRegRes << 24) | (i << 16) | MUL_CMD; // offs0+MA = offsA * offsB+i. mul
        
        gCmd[12 * i + 2] = 0x02020300 | B_SHFT_CMD; // load [C][C][T]
        gCmd[12 * i + 3] = (i << 24) | (i << 16) | ADD_CMD; // offsC+i = offsC+i * offsT. add previous carry
        
        gCmd[12 * i + 4] = 0x03040400 | B_SHFT_CMD; // load [T][0][0]
        gCmd[12 * i + 5] = ((gMulReg + 1) << 8) | MOV_CMD; // offsT = offs0+MA+1. get most significant 128 bits from mul to temp-carry reg
        
        gCmd[12 * i + 6] = 0x03030400 | B_SHFT_CMD; // load [T][T][0]
        gCmd[12 * i + 7] = (gZeroReg << 8) | C_ADD_CMD; // offsT = offsT + offs0+ZR + carry. add carry flag (from last add op) to temp-carry reg
        
        gCmd[12 * i + 8] = 0x02020400 | B_SHFT_CMD; // load [C][C][0]
        gCmd[12 * i + 9] = (i << 24) | (i << 16) | (gMulReg << 8) | ADD_CMD; // offsC+i = offsC+i + offs0+MA. add mul least significant 128 bits to result reg
        
        gCmd[12 * i + 10] = 0x03030400 | B_SHFT_CMD; // load [T][T][0]
        gCmd[12 * i + 11] = (gZeroReg << 8) | C_ADD_CMD; // offsT = offsT + offs0+ZR + carry. add carry flag (from last add op) to temp-carry reg
    }
    
    addCmdToAlu(ALU_Mul, gCmd);
}

////////////////////////////////////////////////////////////////////////////////
// PROCESS CMD
////////////////////////////////////////////////////////////////////////////////
static uint8_t NextSeq = 1;

/**
* Waits while ALU isn't ready
*
* @author Kosenkov
* @date 18.05.2018
*/
void aluProcWait()
{
    uint32_t stat = 0;
    do
    {
        stat = (ALU->CMDC_CTRL << 7) >> 23; // todo: use macro
    } while (stat != 0);
}

/**
* Selects specified ALU commands bank
*
* @param bank ALU commands bank number
* @author Kosenkov
* @date 18.05.2018
*/
void aluSwitchBank(uint8_t bank)
{
    uint32_t stat;
    
    --bank;
    stat = (bank << 31) | ((NextSeq++ & 0x3F) << 10); // todo: use macro
    ALU->CMDC_CTRL = stat;                                // run commands
    aluProcWait();                                        // waiting for this cmd
}

/**
* Process specified command
*
* @param name Of command to process
* @param sz Of complex command in native command's units
* @author Kosenkov
* @date 18.05.2018
*/
void aluProcCmd(Command name, uint32_t sz)
{
    //uint8_t bank = 0; TODO: in current realization we not need 2nd bank
    uint32_t stat;
    
    stat = /*(bank << 31) |*/ (sz << 16) | ((NextSeq++ & 0x3F) << 10) | (gCmds[name].Beg); // todo: use macro
    aluProcWait(); // TODO: we able to not wait everytime. we can push next data when previous is in process
    ALU->CMDC_CTRL = stat;
}

////////////////////////////////////////////////////////////////////////////////
// HELPERS
////////////////////////////////////////////////////////////////////////////////
/**
* Calculates count of most significant zero bits
*
* @param a 128-bit register number
* @param len registers count
* @return zero bits count
* @author Kosenkov
* @date 18.05.2018
*/
uint32_t countZeroBits(Reg128Num a, uint32_t len)
{
    uint32_t res = 0;
    uint32_t i = 0;
    len <<= 2;
    
    while ((ALU_RA + (uint32_t)(a << 2))[len - 1 - i] == 0)
    {
        if (i == len)
        {
            return len * 32;
        }
        ++i;
    }
    res = i;
    i = 0;
    while (((ALU_RA + (uint32_t)(a << 2))[len - 1 - res] & (0x80000000 >> i)) == 0)
    {
        i++;
    }
    return res * 32 + i;
}

////////////////////////////////////////////////////////////////////////////////
// INTERNAL ALU FUNCS
////////////////////////////////////////////////////////////////////////////////
/**
* Sets new bases to base-register
*
* @param res 128-bit register number
* @param a 128-bit register number
* @param b 128-bit register number
* @param tmp 128-bit register number
* @author Kosenkov
* @date 18.05.2018
*/
void aluSetBase(Reg128Num res, Reg128Num a, Reg128Num b, Reg128Num tmp)
{
    ALU->BASE_SHIFT = ((tmp << 24) | (res << 16) | (b << 8) | a);
}

////////////////////////////////////////////////////////////////////////////////
// HEADER IMPL
////////////////////////////////////////////////////////////////////////////////
/**
* ALU initialization
*
* @author Kosenkov
* @date 18.05.2018
*/
void aluInitCCPU()
{
    uint8_t  i;

    for (i = 0; i < CMD_MAX_I; ++i) // TODO: this block uses first bank only
    {
        gCmds[i].Bank = 0;
        if (i == 0)
        {
            gCmds[i].Beg = 0;
        }
        else
        {
            gCmds[i].Beg = gCmds[i - 1].Beg + gCmds[i - 1].Cnt;
        }
        
        if (i == ALU_Push || i == ALU_Pop) // TODO: size independet, more flexible realization
        {
            gCmds[i].Cnt = gStackSz;
        }
        else if (i == ALU_Mul)
        {
            gCmds[i].Cnt = 2 + 12 * gMaxRegSz;
        }
        else
        {
            gCmds[i].Cnt = gMaxRegSz;
        }
    }
    
    aluSwitchBank(0);
    genCmdZero();
    genCmdPush();
    genCmdPop();
    genCmdMov();
    genCmdCMov();
    genCmdAdd();
    genCmdSub();
    genCmdShl();
    genCmdCShl();
    genCmdShr();
    genCmdCShr();
    genCmdMul();
}

/**
* Writes 32-bits word to ALU register
*
* @param num Number of 128-bits ALU register
* @param offs 32-bits offset in 128-bits register
* @param data Word to write
* @author Kosenkov
* @date 18.05.2018
*/
void aluWriteWord(Reg128Num num, Sz32 offs, uint32_t data)
{
    aluProcWait();
    *(ALU_RA + (num << 2) + offs) = data;
}

/**
* Reads 32-bits word from ALU register
*
* @param num Number of 128-bits ALU register
* @param offs 32-bits offset in 128-bits register
* @param data Word to read to
* @author Kosenkov
* @date 18.05.2018
*/
void aluReadWord(Reg128Num num, Sz32 offs, uint32_t* data)
{
    aluProcWait();
    *data = *(ALU_RA + (num << 2) + offs);
}

/**
* Writes array of 32-bits words to ALU register
*
* @param num Number of 128-bits ALU register
* @param sz Size of array in 32-bits words
* @param data Array to write
* @author Kosenkov
* @date 18.05.2018
*/
void aluWriteData(Reg128Num num, Sz32 sz, uint32_t* data)
{
    aluProcWait();
    // TODO: if (sz & 3 != 0)
    writeBy32(ALU_RA + (num << 2), data, sz);
}

/**
* Reads array of 32-bits words from ALU register
*
* @param num Number of 128-bits ALU register
* @param sz Size of array in 32-bits words
* @param data Array to read to
* @author Kosenkov
* @date 18.05.2018
*/
void aluReadData(Reg128Num num, Sz32 sz, uint32_t* data)
{
    aluProcWait();
    // TODO: if (sz & 3 != 0)
    readBy32(ALU_RA + (num << 2), data, sz);
}

/**
* Returns carry flag from ALU
*
* @return CARRY flag if carry bit was set, 0 othrwise
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t aluGetCarry()
{
    aluProcWait();
    return ALU->ALU_SCTRL & CARRY;
}

/**
* Returns shift flag from ALU
*
* @return SHIFT flag if carry bit was set, 0 othrwise
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t aluGetShift()
{
    aluProcWait();
    return ALU->ALU_SCTRL & SHIFT;
}

/**
* Zeroes 128-bits registers
*
* @param res Register to be zeroed
* @param sz Count of registers to be zeroed
* @author Kosenkov
* @date 18.05.2018
*/
void aluZero(Reg128Num res, Sz128 sz)
{
    aluSetBase(res, 0, 0, 0);
    aluProcCmd(ALU_Zero, sz);
}

/**
* Copy data from register a to register res
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluMov(Reg128Num res, Reg128Num a, Sz128 sz)
{
    aluSetBase(res, a, 0, 0);
    aluProcCmd(ALU_Mov, sz);
}

/**
* Move with carry flag usage
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluCMov(Reg128Num res, Reg128Num a, Sz128 sz)
{
    aluSetBase(res, a, 0, 0);
    aluProcCmd(ALU_CMov, sz);
}

/**
* res = a + b
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluAdd(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz)
{
    aluSetBase(res, a, b, 0);
    aluProcCmd(ALU_Add, sz);
}

/**
* res = a - b
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluSub(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz)
{
    aluSetBase(res, a, b, 0);
    aluProcCmd(ALU_Sub, sz);
}

/**
* res = a << 1
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluShl(Reg128Num res, Reg128Num a, Sz128 sz)
{
    aluSetBase(res, a, 0, 0);
    aluProcCmd(ALU_Shl, sz);
}

/**
* res = a << 1 + carry (shift with carry)
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluCShl(Reg128Num res, Reg128Num a, Sz128 sz)
{
    aluSetBase(res, a, 0, 0);
    aluProcCmd(ALU_CShl, sz);
}

/**
* res = a >> 1
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluShr(Reg128Num res, Reg128Num a, Sz128 sz)
{
    aluSetBase(0x7F & (res - gMaxRegSz + sz), 0x7F & (a - gMaxRegSz + sz), 0, 0);
    aluProcCmd(ALU_Shr, sz);
}

/**
* res = a >> 1 + carry (shift with carry)
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluCShr(Reg128Num res, Reg128Num a, Sz128 sz)
{
    aluSetBase(0x7F & (res - gMaxRegSz + sz), 0x7F & (a - gMaxRegSz + sz), 0, 0);
    aluProcCmd(ALU_CShr, sz);
}

/**
* res = a / b
* a = a % b
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param tmp Temporary register
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluDiv(Reg128Num res, Reg128Num a, Reg128Num b, Reg128Num tmp, Sz128 sz)
{
    uint32_t i = 0;
    uint32_t za = 0;
    uint32_t zb = 0;
    uint32_t shfts = 0;
    
    // zero res
    aluZero(res, sz);
    
    za = countZeroBits(a, sz);
    zb = countZeroBits(b, sz);
    if (za > zb)
    {
        return;
    }
    shfts = zb - za;

    // shifts
    aluSetBase(b, b, 0, 0);
    for (i = 0; i < shfts; ++i)
    {
        aluProcCmd(ALU_Shl, sz); // apply if i == 0
    }
    
    // div
    for (i = 0; i < shfts; ++i)
    {
        aluSub(tmp, a, b, sz);
        aluCMov(a, tmp, sz);
        aluCShl(res, res, sz);
        aluShr(b, b, sz);
    }
    aluSub(tmp, a, b, sz);
    aluCMov(a, tmp, sz);
    aluCShl(res, res, sz);
}

/**
* a = a % b
*
* @param a 1st operand
* @param b 2st operand
* @param tmp Temporary register
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluMod(Reg128Num a, Reg128Num b, Reg128Num tmp, Sz128 sz)
{
    uint32_t i = 0;
    uint32_t za = 0;
    uint32_t zb = 0;
    uint32_t shfts = 0;
    
    za = countZeroBits(a, sz);
    zb = countZeroBits(b, sz);
    if (za > zb)
    {
        return;
    }
    shfts = zb - za;

    // shifts
    aluSetBase(b, b, 0, 0);
    for (i = 0; i < shfts; ++i)
    {
        aluProcCmd(ALU_Shl, sz); // apply if i == 0
    }
    
    // mod
    for (i = 0; i < shfts; ++i)
    {
        aluSub(tmp, a, b, sz);
        aluCMov(a, tmp, sz);
        aluShr(b, b, sz);
    }
    aluSub(tmp, a, b, sz);
    aluCMov(a, tmp, sz);
}

/**
* Exchanges data of registers
*
* @param left 1st operand
* @param right 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluXchg(Reg128Num left, Reg128Num right, Sz128 sz)
{
    Reg128Num leftOffs = 0;
    Reg128Num rightOffs = 0;
    
    for (;;)
    {
        if (sz > gStackSz)
        {
            aluSetBase(left + leftOffs, right + rightOffs, left + leftOffs, 0);
            aluProcCmd(ALU_Push, gStackSz);
            aluSetBase(right + rightOffs, 0, 0, 0);
            aluProcCmd(ALU_Pop, gStackSz);
            sz -= gStackSz;
            leftOffs += gStackSz;
            rightOffs += gStackSz;
        }
        else
        {
            aluSetBase(left + leftOffs, right + rightOffs, left + leftOffs, 0);
            aluProcCmd(ALU_Push, sz);
            aluSetBase(0x7F & (right + rightOffs - gStackSz + sz), 0, 0, 0);
            aluProcCmd(ALU_Pop, sz);
            return;
        }
    }
}

/**
* res = a * b. res is doublesized
*
* @param res Resulting register (its size is sz * 2)
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluMul(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz)
{
    Sz128 i;
    Sz128 resSz = sz << 1;

    aluZero(res, resSz);
    for (i = 0; i < sz; ++i)
    {
        aluSetBase(res + i, a + i, b, res + sz + i);
        aluProcCmd(ALU_Mul, 12 * sz);
    }
}

/**
* res = [least significant half of] a * b. res is singlesized
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluHalfMul(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz)
{
    Sz128 i;

    aluZero(res, sz);
    aluSetBase(111, gZeroReg, 111, 0);
    aluProcCmd(ALU_Push, 1);
    for (i = 0; i < sz; ++i)
    {
        aluSetBase(res + i, a + i, b, 111);
        aluProcCmd(ALU_Mul, 12 * (sz - i));
        aluSetBase(111, 0, 0, 0);
        aluProcCmd(ALU_Zero, 1);
    }
    aluSetBase(111 - (gStackSz - 1), 0, 0, 0);
    aluProcCmd(ALU_Pop, 1);
}
