/**
* @file gate.c
* @brief Gate unit implementation
* @author Kosenkov
* @date 01.02.2018
*/

#include "cm0.h"
#include "gate.h"

/**
* Pushes 32-bits word to gate stack
*
* @param gate Gate number
* @param word To push
* @author Kosenkov
* @date 18.05.2018
*/
void gatePushWord(Gates gate, uint32_t word)
{
    GATE_TypeDef* gates[2] = {GATE0_REGS, GATE1_REGS};
    for (;;)
    {
        if (gates[gate]->SF_FIFO_INT_SOURCE & FLAG_SF_FIFO_IN_NOT_FULL)
        {   // break and write if fifo is not full
            break;
        }
    }
    gates[gate]->SF_FIFO_SF_TO_OP = word;
}

/**
* Reads 32-bits word from gate stack
*
* @param gate Gate number
* @param word To read to
* @author Kosenkov
* @date 18.05.2018
*/
void gatePopWord(Gates gate, uint32_t* word)
{
    GATE_TypeDef* gates[2] = {GATE0_REGS, GATE1_REGS};
    for (;;)
    {
        if (gates[gate]->SF_FIFO_INT_SOURCE & FLAG_SF_FIFO_OUT_NOT_EMPTY)
        {   // break and read if fifo is not empty
            break;
        }
    }
    
    *word = gates[gate]->SF_FIFO_OP_TO_SF;
}

/**
* Pushes 32-bits words array to gate stack
*
* @param gate Gate number
* @param data Array to push
* @param sz Size of array
* @author Kosenkov
* @date 18.05.2018
*/
void gatePushData(Gates gate, uint32_t* data, Sz32 sz)
{
    uint32_t i;
    for (i = 0; i < sz; ++i)
    {
        gatePushWord(gate, data[i]);
    }
}

/**
* Reads 32-bits words array from gate stack
*
* @param gate Gate number
* @param data Array to read from stack
* @param sz Size of array
* @author Kosenkov
* @date 18.05.2018
*/
void gatePopData(Gates gate, uint32_t* data, Sz32 sz)
{
    uint32_t i;
    for (i = 0; i < sz; ++i)
    {
        gatePopWord(gate, &data[i]);
    }
}

/**
* Set gate register to specified state
*
* @param gate Gate number
* @param regNum Register number
* @param word To store in register
* @author Kosenkov
* @date 18.05.2018
*/
void gateSetReg(Gates gate, uint8_t regNum, uint32_t word)
{
    GATE_TypeDef* gates[2] = {GATE0_REGS, GATE1_REGS};
    gates[gate]->SF_REG[regNum] = word;
}

/**
* Get gate register state
*
* @param gate Gate number
* @param regNum Register number
* @param word To read from register
* @author Kosenkov
* @date 18.05.2018
*/
void gateGetReg(Gates gate, uint8_t regNum, uint32_t* word)
{
    GATE_TypeDef* gates[2] = {GATE0_REGS, GATE1_REGS};
    *word = gates[gate]->SF_REG[regNum];
}
