/**
* @file otp.c
* @brief OTP access implementation
* @author Kosenkov
* @date 18.05.2018
*/

#include "cm0.h"
#include "otp.h"

/**
* Reads data from OTP
*
* @param mem Address to write to
* @param otp Address in OTP to read from
* @param sz In 32-bits units
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 18.05.2018
*/
CM0_STATUS otpRead(uint32_t* mem, const uint32_t* otp, Sz32 sz)
{
    uint32_t i;
    uint32_t bitAddr = ((uint32_t)otp - (uint32_t)OTP_MEM) << 1; // (bytes -> bits) / 4 banks
    
    for (i = 0; i < sz; ++i)
    {
        OTP->RW_CMD_REG = (bitAddr + (i << 3)) | OTP_CMD_READ;
        while (OTP->STAT_CTRL_REG & OTP_BUSY) // wait while busy
        {}
        if (OTP->RW_CMD_REG & OTP_CMD_RW_ERROR)
        {
            return CM0_FAIL;
        }
        mem[i] = OTP->READ_DATA_REG;
    }
    
    return CM0_OK;
}

/**
* Writes data to OTP
*
* @param mem Address to read from
* @param otp Address in OTP to write to
* @param sz In 32-bits units
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 18.05.2018
*/
CM0_STATUS otpWrite(const uint32_t* otp, uint32_t* mem, Sz32 sz)
{
    uint32_t i;
    uint32_t j;

    uint32_t step;
    uint32_t word;
    uint8_t* word8 = (uint8_t*)&word;
    const uint32_t mask = 0x01010101;
    uint32_t checkWord = 0xFFFFFFFF;
    uint32_t bitAddr = ((uint32_t)otp - (uint32_t)OTP_MEM) << 1; // (bytes -> bits) / 4 banks
    
    for (i = 0; i < sz; ++i)
    {
        step = mem[i];
        for (j = 0; j < 8; ++j)
        {
            // write
            word = mask & step & checkWord;
            OTP->RW_CMD_REG = ((bitAddr + j) | (word8[0] << 24) | (word8[1] << 25) | (word8[2] << 26) | (word8[3] << 27) | OTP_CMD_WRITE);
            while ((OTP->STAT_CTRL_REG & OTP_BUSY) != 0) // while busy
            {}
            if ((OTP->RW_CMD_REG & OTP_CMD_RW_ERROR) != 0)
            {
                return CM0_FAIL;
            }
            
            // read and check
            OTP->RW_CMD_REG = bitAddr | OTP_CMD_READ;
            while (OTP->STAT_CTRL_REG & OTP_BUSY) // wait while busy
            {}
            if (OTP->RW_CMD_REG & OTP_CMD_RW_ERROR)
            {
                return CM0_FAIL;
            }
            checkWord = (OTP->READ_DATA_REG >> j) & mask & checkWord;
            if (checkWord != word) // rollback changes and try again
            {
                --j;
                checkWord = ~checkWord & word; // get bits that weren't written
                continue;
            }
            checkWord = 0xFFFFFFFF;
            step >>= 1;
        }
        bitAddr += 8;
    }
    
    return CM0_OK;
}
