/**
* @file rng.c
* @brief Random generator implementation
* @author Kosenkov
* @date 18.04.2018
*/

#include "rng.h"
#include "cm0.h"

#include "types.h"
#include "rngTest.h"

#define RNG_CLK_DESIRED ((uint32_t)500000) /*< 500 KHz */

/**
* Sets to random generator current core clock
* It is important for proper RNG working
*
* @param clock Current core clock
* @author Kosenkov
* @date 18.05.2018
*/
void rngSetClock(uint32_t clock)
{
    RNG->CLK_DIV = (((clock / RNG_CLK_DESIRED) >> 1) - 1);
}

/**
* Sets random generator mode
* @see cm0.h
*
* @param mode Work mode
* @author Kosenkov
* @date 18.05.2018
*/
void rngSetMode(uint32_t mode)
{
    RNG->STAT_CTRL = (RNG->STAT_CTRL & ~RNG_WORK_MODE_MASK) | (mode & RNG_WORK_MODE_MASK);
    if (mode != RNG_WORK_MODE_STOPPED) // waiting for RNG module
    {
        while ((RNG->STAT_CTRL & (RNG_UNREAD_DATA | RNG_BUSY)) == 0)
        {}
    }
}

/**
* Gets random number
*
* @param word Address for random number store to
* @retval CM0_OK if success
* @retval CM0_FAIL otherwise
* @author Kosenkov
* @date 18.05.2018
*/
CM0_STATUS rngGetWord(uint32_t* word)
{
    uint32_t status = 0;
    
    while (RNG->PAUSE_CNT != 0) // pause
    {}
    
    for (;;)
    {
        status = RNG->STAT_CTRL;
        if (status & RNG_UNREAD_DATA) // can get word
        {
            *word = RNG->OUTPUT;
            return CM0_OK;
        }
        else if (status & RNG_BUSY) // working
        {
            continue;
        }
        else
        {
            break;
        }
    }
    return CM0_FAIL; // random number isn't ready, generator isn't busy
}

/**
* Runs random quality tests
* 
* @param clock Current core clock
* @retval CM0_OK if success
* @retval CM0_RNG_TEST_ERR otherwise
* @author Kosenkov
* @date 23.05.2018
*/
CM0_STATUS rngTest(uint32_t clock)
{
    const Sz32 sz = STAT_CNT / sizeof(uint32_t);
    uint8_t random[sizeof(uint32_t)];
    uint8_t i;
    uint16_t j;
    uint8_t k;
    
    // clock
    rngSetClock(clock);
    
    // init statistical tests
    rngTestInit();
    
    for (i = 0; i < 3; ++i)
    {
        // set working mode and gen randoms
        rngSetMode(RNG_WORK_MODE_AFTER_READ);
        for (j = 0; j < sz; ++j) // collect data and test
        {
            if (CM0_CHK_ERR(rngGetWord((uint32_t*)random)))
            {
                rngSetMode(RNG_WORK_MODE_STOPPED);
                return CM0_FAIL;
            }
            
            // test bytes
            for (k = 0; k < sizeof(uint32_t); ++k)
            {
                rngTestByte(random[k]);
            }
        }
        // stop random module
        rngSetMode(RNG_WORK_MODE_STOPPED);
        
        // check results
        if (CM0_CHK_ERR(rngTestResult()))
        {
            continue;
        }
        else
        {
            return CM0_OK;
        }
    }

    return CM0_RNG_TEST_ERR;
}
